
from .base import TestBase

from pymap.imap import IMAPServer


class TestMailbox(TestBase):

    async def test_rename(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'status1 STATUS Sent (MESSAGES UIDNEXT UIDVALIDITY)\r\n')
        transport.push_write(
            b'* STATUS Sent (MESSAGES 2 UIDNEXT 103 '
            b'UIDVALIDITY ', (br'\d+', b'uidval1'), b')\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_readline(
            b'rename1 RENAME Sent "Sent Test"\r\n')
        transport.push_write(
            b'rename1 OK RENAME completed.\r\n')
        transport.push_readline(
            b'status1 STATUS Sent (MESSAGES)\r\n')
        transport.push_write(
            b'status1 NO [NONEXISTENT] Mailbox does not exist.\r\n')
        transport.push_readline(
            b'status1 STATUS "Sent Test" (MESSAGES UIDNEXT UIDVALIDITY)\r\n')
        transport.push_write(
            b'* STATUS "Sent Test" (MESSAGES 2 UIDNEXT 103 '
            b'UIDVALIDITY ', (br'\d+', b'uidval2'), b')\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['uidval1'] == self.matches['uidval2']

    async def test_rename_inbox(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'status1 STATUS INBOX (MESSAGES UIDNEXT UIDVALIDITY)\r\n')
        transport.push_write(
            b'* STATUS INBOX (MESSAGES 4 UIDNEXT 105 '
            b'UIDVALIDITY ', (br'\d+', b'uidval1'), b')\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_readline(
            b'rename1 RENAME INBOX "Inbox Test"\r\n')
        transport.push_write(
            b'rename1 OK RENAME completed.\r\n')
        transport.push_readline(
            b'status1 STATUS INBOX (MESSAGES UIDNEXT UIDVALIDITY)\r\n')
        transport.push_write(
            b'* STATUS INBOX (MESSAGES 0 UIDNEXT 101 '
            b'UIDVALIDITY ', (br'\d+', b'uidval2'), b')\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_readline(
            b'status1 STATUS "Inbox Test" (MESSAGES UIDNEXT UIDVALIDITY)\r\n')
        transport.push_write(
            b'* STATUS "Inbox Test" (MESSAGES 4 UIDNEXT 105 '
            b'UIDVALIDITY ', (br'\d+', b'uidval3'), b')\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['uidval1'] != self.matches['uidval2']
        assert self.matches['uidval1'] == self.matches['uidval3']

    async def test_rename_inbox_selected(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'rename1 RENAME INBOX "Inbox Test"\r\n')
        transport.push_write(
            b'rename1 OK RENAME completed.\r\n')
        transport.push_select(b'INBOX', 0, 0, 101, False)
        transport.push_logout()
        await self.run(transport)

    async def test_rename_other_selected(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'rename1 RENAME Sent "Sent Test"\r\n')
        transport.push_write(
            b'* BYE Selected mailbox no longer exists.\r\n'
            b'rename1 OK RENAME completed.\r\n')
        await self.run(transport)

    async def test_rename_selected(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'rename1 RENAME Sent "Sent Test"\r\n')
        transport.push_write(
            b'* BYE Selected mailbox no longer exists.\r\n'
            b'rename1 OK RENAME completed.\r\n')
        await self.run(transport)

    async def test_rename_inferior(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'create1 CREATE Test\r\n')
        transport.push_write(
            b'create1 OK [MAILBOXID (', (br'F[a-f0-9]+', ), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'create2 CREATE Test/One\r\n')
        transport.push_write(
            b'create2 OK [MAILBOXID (', (br'F[a-f0-9]+', ), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'create3 CREATE Test/One/Two\r\n')
        transport.push_write(
            b'create3 OK [MAILBOXID (', (br'F[a-f0-9]+', ), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'delete1 DELETE Test/One\r\n')
        transport.push_write(
            b'delete1 OK DELETE completed.\r\n')
        transport.push_readline(
            b'rename1 RENAME Test Foo\r\n')
        transport.push_write(
            b'rename1 OK RENAME completed.\r\n')
        transport.push_readline(
            b'list1 LIST Test *\r\n')
        transport.push_write(
            b'list1 OK LIST completed.\r\n')
        transport.push_readline(
            b'list2 LIST Foo *\r\n')
        transport.push_write(
            b'* LIST (\\HasChildren) "/" Foo\r\n'
            b'* LIST (\\Noselect \\HasChildren) "/" Foo/One\r\n'
            b'* LIST (\\HasNoChildren) "/" Foo/One/Two\r\n'
            b'list2 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_rename_mailbox_id(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'create1 CREATE Test\r\n')
        transport.push_write(
            b'create1 OK [MAILBOXID (', (br'F[a-f0-9]+', b'mbxid'), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'rename1 RENAME Test Foo\r\n')
        transport.push_write(
            b'rename1 OK RENAME completed.\r\n')
        transport.push_select(b'Foo', unseen=False)
        transport.push_logout()
        await self.run(transport)
        assert self.matches['mbxid1'] == self.matches['mbxid']




import asyncio
import enum
import inspect
import re
import socket
import traceback
from collections import deque
from itertools import zip_longest
from typing import overload, Any, Literal, NoReturn

from pymap.concurrent import Event
from pymap.imap import IMAPServer
from pymap.sieve.manage import ManageSieveServer

__all__ = ['MockTransport']


class _Type(enum.Enum):
    READLINE = "reader.readline()"
    READEXACTLY = "reader.readexactly()"
    DRAIN = "writer.drain()"
    READ_EOF = "reader.at_eof()"


class _Socket:

    def __init__(self, fd: int) -> None:
        self.fd = fd
        self.family = socket.AF_INET

    def fileno(self) -> int:
        return self.fd


_WriteDataTuple = tuple['_WriteDataPart', ...]
_WriteDataPart = bytes | int | _WriteDataTuple | None
_WriteData = bytes | int | _WriteDataTuple
_ReadLineOp = tuple[Literal[_Type.READLINE], str, bytes,
                    Event | None, Event | None]
_ReadExactlyOp = tuple[Literal[_Type.READEXACTLY], str, bytes,
                       Event | None, Event | None]
_DrainOp = tuple[Literal[_Type.DRAIN], str, _WriteDataTuple,
                 Event | None, Event | None]
_ReadEofOp = tuple[Literal[_Type.READ_EOF], str, None,
                   Event | None, Event | None]
_Operation = _ReadLineOp | _ReadExactlyOp | _DrainOp | _ReadEofOp
_Server = IMAPServer | ManageSieveServer


class MockTransport:

    def __init__(self, server: _Server, matches: dict[str, bytes], fd) -> None:
        self.server = server
        self.queue: deque[_Operation] = deque()
        self.matches = matches
        self.socket = _Socket(fd)
        self._write_batch: list[bytes] = []
        self._select_count = 0

    @classmethod
    def _caller(cls, frame) -> str:
        frame = frame.f_back if frame else None
        fields = inspect.getframeinfo(frame) if frame else ('?', '?')
        return '{0}:{1!s}'.format(fields[0], fields[1])

    @classmethod
    def _fail(cls, msg: str) -> NoReturn:
        raise AssertionError(msg)

    def push_readline(self, data: bytes, wait: Event | None = None,
                      set: Event | None = None) -> None:
        where = self._caller(inspect.currentframe())
        self.queue.append((_Type.READLINE, where, data, wait, set))

    def push_readexactly(self, data: bytes, wait: Event | None = None,
                         set: Event | None = None) -> None:
        where = self._caller(inspect.currentframe())
        self.queue.append((_Type.READEXACTLY, where, data, wait, set))

    def push_write(self, *data: _WriteDataPart, wait: Event | None = None,
                   set: Event | None = None) -> None:
        where = self._caller(inspect.currentframe())
        self.queue.append((_Type.DRAIN, where, data, wait, set))

    def push_read_eof(self, wait: Event | None = None,
                      set: Event | None = None) -> None:
        where = self._caller(inspect.currentframe())
        self.queue.append((_Type.READ_EOF, where, None, wait, set))

    def push_login(self, password: bytes = b'testpass',
                   wait: Event | None = None,
                   set: Event | None = None) -> None:
        self.push_write(
            b'* OK [CAPABILITY IMAP4rev1',
            (br'(?:\s+[a-zA-Z0-9=+-]+)*', ),
            b'] Server ready ',
            (br'\S+', ), b'\r\n', wait=wait)
        self.push_readline(
            b'login1 LOGIN testuser ' + password + b'\r\n')
        self.push_write(
            b'login1 OK [CAPABILITY IMAP4rev1',
            (br'(?:\s+[a-zA-Z0-9=+-]+)*', ),
            b'] Authentication successful.\r\n', set=set)

    def push_logout(self, wait: Event | None = None, set: Event | None = None):
        self.push_readline(
            b'logout1 LOGOUT\r\n', wait=wait)
        self.push_write(
            b'* BYE Logging out.\r\n'
            b'logout1 OK Logout successful.\r\n', set=set)

    def push_select(self, mailbox: bytes,
                    exists: _WriteData | None = None,
                    recent: _WriteData | None = None,
                    uidnext: _WriteData | None = None,
                    unseen: _WriteData | Literal[False] | None = None,
                    readonly: bool = False, examine: bool = False,
                    wait: Event | None = None,
                    post_wait: Event | None = None,
                    set: Event | None = None) -> None:
        n = self._select_count = self._select_count + 1
        if unseen is False:
            unseen_line: _WriteData = (None, b'')
        elif unseen is None:
            unseen_line = (None, b'* OK [UNSEEN ', (br'\d+',),
                           b'] First unseen message.\r\n')
        else:
            unseen_line = (None, b'* OK [UNSEEN ', unseen,
                           b'] First unseen message.\r\n')
        if exists is None:
            exists = (br'\d+', )
        if recent is None:
            recent = (br'\d+', )
        if uidnext is None:
            uidnext = (br'\d+', )
        if readonly or examine:
            ok_code = b'READ-ONLY'
            permflags_line = b'* OK [PERMANENTFLAGS ()] Read-only mailbox.\r\n'
        else:
            ok_code = b'READ-WRITE'
            permflags_line = b'* OK [PERMANENTFLAGS (' \
                             b'\\Answered \\Deleted \\Draft \\Flagged ' \
                             b'\\Seen)] Flags permitted.\r\n'
        if examine:
            tag = b'examine%i' % (n, )
            cmd = b'EXAMINE'
        else:
            tag = b'select%i' % (n, )
            cmd = b'SELECT'
        self.push_readline(
            tag + b' ' + cmd + b' ' + mailbox + b'\r\n', wait=wait)
        self.push_write(
            permflags_line,
            b'* FLAGS (\\Answered \\Deleted \\Draft '
            b'\\Flagged \\Recent \\Seen)\r\n'
            b'* ', exists, b' EXISTS\r\n'
            b'* ', recent, b' RECENT\r\n'
            b'* OK [UIDNEXT ', uidnext, b'] Predicted next UID.\r\n'
            b'* OK [UIDVALIDITY ', (br'\d+', ), b'] UIDs valid.\r\n',
            unseen_line,
            b'* OK [MAILBOXID (', (br'F[a-f0-9]+', b'mbxid%i' % (n, )), b')] '
            b'Object ID.\r\n',
            tag, b' OK [', ok_code, b'] Selected mailbox.\r\n',
            wait=post_wait, set=set)

    @overload
    def _pop_expected(self, got: Literal[_Type.READLINE]) -> _ReadLineOp:
        ...

    @overload
    def _pop_expected(self, got: Literal[_Type.READEXACTLY]) -> _ReadExactlyOp:
        ...

    @overload
    def _pop_expected(self, got: Literal[_Type.DRAIN]) -> _DrainOp:
        ...

    @overload
    def _pop_expected(self, got: Literal[_Type.READ_EOF]) -> _ReadEofOp:
        ...

    def _pop_expected(self, got: _Type) -> _Operation:
        try:
            try:
                type_, where, data, wait, set = self.queue.popleft()
            except IndexError as exc:
                raise AssertionError('\nExpected: <end>'
                                     '\nGot:      ' + got.value) from exc
            if type_ != got:
                raise AssertionError('\nExpected: ' + type_.value +
                                     '\nGot:      ' + got.value +
                                     '\nWhere:    ' + where)
        except AssertionError:
            traceback.print_exc()
            raise
        return got, where, data, wait, set  # type: ignore

    def _match_write_expected(self, expected: _WriteDataTuple,
                              re_parts: list[bytes]) -> None:
        for part in expected:
            if part is None:
                re_parts.append(br'.*?')
            elif isinstance(part, bytes):
                re_parts.append(re.escape(part))
            elif isinstance(part, int):
                re_parts.append(b'%i' % part)
            else:
                if len(part) == 1 or part[1] is None:
                    assert isinstance(part[0], bytes)
                    re_parts.append(part[0])
                elif part[0] is None:
                    self._match_write_expected(part[1:], re_parts)
                else:
                    regex, name = part
                    assert isinstance(regex, bytes)
                    assert isinstance(name, bytes)
                    re_parts.append(br'(?P<' + name + br'>' + regex + br')')

    def _match_write_msg(self, expected: _WriteData, data: bytes,
                         full_regex: bytes, where: str) -> str:
        parts = ['',
                 'Expected: ' + repr(expected),
                 'Got:      ' + repr((data, )),
                 'Where:    ' + where,
                 '---------------------------------------------------']
        raw_regex = str(full_regex.replace(b'\r', b''), 'utf-8').splitlines()
        raw_data = str(data, 'utf-8').splitlines()
        for regex_line, data_line in zip_longest(raw_regex, raw_data):
            if regex_line:
                parts.append('regex: ' + regex_line.lstrip('^'))
            else:
                parts.append('regex: ')
            if data_line:
                parts.append('data:  ' + re.escape(data_line + '\\'))
            else:
                parts.append('data:  ')
            parts.append('')
        return '\n'.join(parts)

    def _match_write(self, where: str, expected: _WriteDataTuple,
                     data: bytes) -> None:
        re_parts: list[bytes] = []
        self._match_write_expected(expected, re_parts)
        full_regex = b'^' + b''.join(re_parts) + b'$'
        match = re.search(full_regex, data)
        if not match:
            raise AssertionError(
                self._match_write_msg(expected, data, full_regex, where))
        self.matches.update(match.groupdict())

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        if name == 'socket':
            return self.socket
        elif name == 'peername':
            return ('1.2.3.4', 1234)
        elif name == 'sockname':
            return ('5.6.7.8', 5678)

    async def readline(self) -> bytes:
        _, where, data, wait, set = self._pop_expected(_Type.READLINE)
        if set:
            set.set()
        if wait:
            try:
                await asyncio.wait_for(wait.wait(), timeout=1.0)
            except asyncio.TimeoutError:
                self._fail('\nTimeout: 1.0s' +
                           '\nWhere:   ' + where)
        return data % {key.encode('ascii'): val
                       for key, val in self.matches.items()}

    async def readexactly(self, size: int) -> bytes:
        _, where, data, wait, set = self._pop_expected(_Type.READEXACTLY)
        if size != len(data):
            raise AssertionError('\nExpected: ' + repr(len(data)) +
                                 '\nGot:      ' + repr(size) +
                                 '\nWhere:    ' + where)
        if set:
            set.set()
        if wait:
            try:
                await asyncio.wait_for(wait.wait(), timeout=1.0)
            except asyncio.TimeoutError:
                self._fail('\nTimeout: 1.0s' +
                           '\nWhere:   ' + where)
        return data

    def write(self, data: bytes) -> None:
        self._write_batch.append(data)

    async def drain(self) -> None:
        _, where, expected, wait, set = self._pop_expected(_Type.DRAIN)
        data = b''.join(self._write_batch)
        self._write_batch = []
        self._match_write(where, expected, data)
        if set:
            set.set()
        if wait:
            try:
                await asyncio.wait_for(wait.wait(), timeout=1.0)
            except asyncio.TimeoutError:
                self._fail('\nTimeout: 1.0s')

    def at_eof(self) -> Literal[False]:
        return False

    def close(self) -> None:
        pass




from .base import TestBase

from pymap.imap import IMAPServer


class TestCopy(TestBase):

    async def test_copy(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'copy1 COPY 1:* Sent\r\n')
        transport.push_write(
            b'copy1 OK [COPYUID ', (br'\d+', ), b' 101:104 103:106]'
            b' COPY completed.\r\n')
        transport.push_select(b'Sent', 6, 4)
        transport.push_logout()
        await self.run(transport)

    async def test_uid_copy(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'copy1 UID COPY 101:* Sent\r\n')
        transport.push_write(
            b'copy1 OK [COPYUID ', (br'\d+', ), b' 101:104 103:106]'
            b' UID COPY completed.\r\n')
        transport.push_select(b'Sent', 6, 4)
        transport.push_logout()
        await self.run(transport)

    async def test_copy_email_id(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'copy1 COPY 4 INBOX\r\n')
        transport.push_write(
            b'* 5 EXISTS\r\n'
            b'* 2 RECENT\r\n'
            b'* 5 FETCH (FLAGS (\\Recent))\r\n'
            b'copy1 OK [COPYUID ', (br'\d+', ), b' 104 105]'
            b' COPY completed.\r\n')
        transport.push_readline(
            b'fetch1 FETCH 4,5 (EMAILID THREADID)\r\n')
        transport.push_write(
            b'* 4 FETCH (EMAILID (', (br'M[a-f0-9]+', b'mid1'), b')'
            b' THREADID (', (br'T[a-f0-9]+', b'tid1'), b'))\r\n'
            b'* 5 FETCH (EMAILID (', (br'M[a-f0-9]+', b'mid2'), b')'
            b' THREADID (', (br'T[a-f0-9]+', b'tid2'), b'))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['mid1'] == self.matches['mid2']
        assert self.matches['tid1'] == self.matches['tid2']

    async def test_concurrent_copy_fetch(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        concurrent = self.new_transport(imap_server)
        event1, event2, event3 = self.new_events(3)

        concurrent.push_login()
        concurrent.push_select(b'Sent', 2, 0, set=event1)
        concurrent.push_readline(
            b'noop1 NOOP\r\n', wait=event2)
        concurrent.push_write(
            b'* 6 EXISTS\r\n'
            b'* 4 RECENT\r\n'
            b'* 3 FETCH (FLAGS (\\Recent \\Seen))\r\n'
            b'* 4 FETCH (FLAGS (\\Answered \\Recent \\Seen))\r\n'
            b'* 5 FETCH (FLAGS (\\Flagged \\Recent))\r\n'
            b'* 6 FETCH (FLAGS (\\Recent))\r\n'
            b'noop1 OK NOOP completed.\r\n')
        concurrent.push_logout()

        transport.push_login()
        transport.push_select(b'INBOX', wait=event1, set=event2)
        transport.push_readline(
            b'copy1 COPY 1:* Sent\r\n')
        transport.push_write(
            b'copy1 OK [COPYUID ', (br'\d+', ), b' 101:104 103:106]'
            b' COPY completed.\r\n', set=event2)
        transport.push_logout()

        await self.run(transport, concurrent)

    async def test_move(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'move1 MOVE 1:* Sent\r\n')
        transport.push_write(
            b'* OK [COPYUID ', (br'\d+', ), b' 101:104 103:106] Moved.\r\n'
            b'* 4 EXPUNGE\r\n'
            b'* 3 EXPUNGE\r\n'
            b'* 2 EXPUNGE\r\n'
            b'* 1 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'move1 OK MOVE completed.\r\n')
        transport.push_select(b'Sent', 6, 4)
        transport.push_logout()
        await self.run(transport)

    async def test_uid_move(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'move1 UID MOVE 101:* Sent\r\n')
        transport.push_write(
            b'* OK [COPYUID ', (br'\d+', ), b' 101:104 103:106] Moved.\r\n'
            b'* 4 EXPUNGE\r\n'
            b'* 3 EXPUNGE\r\n'
            b'* 2 EXPUNGE\r\n'
            b'* 1 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'move1 OK UID MOVE completed.\r\n')
        transport.push_select(b'Sent', 6, 4)
        transport.push_logout()
        await self.run(transport)

    async def test_move_email_id(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 4 (EMAILID THREADID)\r\n')
        transport.push_write(
            b'* 4 FETCH (EMAILID (', (br'M[a-f0-9]+', b'mid1'), b')'
            b' THREADID (', (br'T[a-f0-9]+', b'tid1'), b'))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'move1 MOVE 4 Sent\r\n')
        transport.push_write(
            b'* OK [COPYUID ', (br'\d+', ), b' 104 103] Moved.\r\n'
            b'* 4 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'move1 OK MOVE completed.\r\n')
        transport.push_select(b'Sent', 3, 1)
        transport.push_readline(
            b'fetch2 FETCH 3 (EMAILID THREADID)\r\n')
        transport.push_write(
            b'* 3 FETCH (EMAILID (', (br'M[a-f0-9]+', b'mid2'), b')'
            b' THREADID (', (br'T[a-f0-9]+', b'tid2'), b'))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['mid1'] == self.matches['mid2']
        assert self.matches['tid1'] == self.matches['tid2']

    async def test_concurrent_move_fetch(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        concurrent = self.new_transport(imap_server)
        event1, event2, event3 = self.new_events(3)

        concurrent.push_login()
        concurrent.push_select(b'Sent', 2, 0, set=event1)
        concurrent.push_readline(
            b'noop1 NOOP\r\n', wait=event2)
        concurrent.push_write(
            b'* 4 EXISTS\r\n'
            b'* 2 RECENT\r\n'
            b'* 3 FETCH (FLAGS (\\Recent \\Seen))\r\n'
            b'* 4 FETCH (FLAGS (\\Recent))\r\n'
            b'noop1 OK NOOP completed.\r\n')
        concurrent.push_logout()

        transport.push_login()
        transport.push_select(b'INBOX', wait=event1, set=event2)
        transport.push_readline(
            b'move1 MOVE 1,4 Sent\r\n')
        transport.push_write(
            b'* OK [COPYUID ', (br'\d+', ), b' 101,104 103:104] Moved.\r\n'
            b'* 4 EXPUNGE\r\n'
            b'* 1 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'move1 OK MOVE completed.\r\n', set=event2)
        transport.push_logout()

        await self.run(transport, concurrent)




from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping, MutableSet
from typing import Final

from pysasl.creds.server import ServerCredentials
from pysasl.hashing import Cleartext
from pysasl.identity import Identity

from .config import IMAPConfig
from .exceptions import InvalidAuth

__all__ = ['UserRoles', 'UserMetadata']


class UserRoles(MutableSet[str]):
    """A set of roles for a user.

    Args:
        roles: Initial roles for the set.

    """

    __slots__ = ['_roles']

    def __init__(self, /, roles: Iterable[str | None] = ()) -> None:
        super().__init__()
        self._roles: set[str] = set(role for role in roles if role is not None)

    def __contains__(self, role: object) -> bool:
        return role in self._roles

    def __iter__(self) -> Iterator[str]:
        return iter(self._roles)

    def __len__(self) -> int:
        return len(self._roles)

    def add(self, role: str | None) -> None:
        if role is not None:
            self._roles.add(role)

    def discard(self, role: str | None) -> None:
        if role is not None:
            self._roles.discard(role)


class UserMetadata(Identity):
    """Contains user metadata such as the password or hash.

    Args:
        config: The configuration object.
        authcid: The authentication identity for the user.
        password: The password string or hash digest.
        params: The user metadata parameters.

    """

    _empty_params: Mapping[str, str] = {}

    def __init__(self, config: IMAPConfig, authcid: str, *,
                 password: str | None = None,
                 params: Mapping[str, str] | None = None) -> None:
        super().__init__()
        self.config: Final = config
        self._authcid = authcid
        self._password = password
        self._params = params or {}

    @classmethod
    async def create(cls, config: IMAPConfig, authcid: str, *,
                     password: str | None = None,
                     params: Mapping[str, str] | None = None) -> UserMetadata:
        """Create a new :class:`UserMetadata` by hashing the given *password*
        using the configured hash algorithm. Uses the
        :attr:`~pymap.config.cpu_subsystem` to avoid blocking the main thread.

        Args:
            config: The configuration object.
            authcid: The authentication identity for the user.
            password: The password string or hash digest.
            params: The user metadata parameters.

        """
        if password is not None:
            fut = cls._hash(config, password)
            password = await config.cpu_subsystem.execute(fut)
        return cls(config, authcid, password=password, params=params)

    @classmethod
    async def _hash(cls, config: IMAPConfig, password: str) -> str:
        prepped = config.password_prep(password)
        return config.hash_context.hash(prepped)

    @property
    def authcid(self) -> str:
        return self._authcid

    def compare_secret(self, value: str) -> bool:
        password = self.password
        if password is not None:
            hash_context = self.config.hash_context.copy()
            return hash_context.verify(value, password)
        return False

    def get_clear_secret(self) -> str | None:
        if isinstance(self.config.hash_context, Cleartext):
            return self.password
        return None

    @property
    def password(self) -> str | None:
        """The password string or hash digest."""
        return self._password

    @property
    def roles(self) -> UserRoles:
        """A set of role strings given to the user. These strings only have
        meaning to the backend.

        """
        return UserRoles()

    @property
    def params(self) -> Mapping[str, str]:
        """Additional parameters associated with the user metadata."""
        return self._params

    def get_key(self, identifier: str) -> bytes | None:
        """Find the token key for the given identifier associated with this
        user.

        Args:
            identifier: Any string that can facilitate the lookup of the key.

        """
        return None

    async def check_password(self, creds: ServerCredentials) -> None:
        """Check the given credentials against the known password comparison
        data. If the known data used a hash, then the equivalent hash of the
        provided secret is compared.

        Args:
            creds: The provided authentication credentials.
            token_key: The token key bytestring.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        cpu_subsystem = self.config.cpu_subsystem
        fut = self._check_secret(creds)
        if not await cpu_subsystem.execute(fut):
            raise InvalidAuth()

    async def _check_secret(self, creds: ServerCredentials) -> bool:
        return creds.verify(self)




from __future__ import annotations

from collections.abc import Sequence, Set
from datetime import datetime
from functools import cached_property
from typing import Final

from ..config import IMAPConfig
from ..interfaces.token import TokenCredentials, TokensInterface
from ..plugin import Plugin

__all__ = ['tokens', 'TokensBase', 'AllTokens']

#: Registers token plugins.
tokens: Plugin[TokensBase] = Plugin('pymap.token', default='macaroon')


class TokensBase(TokensInterface):
    """Base class for token types registered by :data:`tokens`.

    Args:
        config: The IMAP configuration object.

    """

    def __init__(self, config: IMAPConfig) -> None:
        super().__init__()
        self.config: Final = config


class AllTokens(TokensBase):
    """Uses :data:`tokens` to support all registered token types.

    For token creation, the :attr:`~pymap.plugin.Plugin.default` token plugin
    is used. For token parsing, each token plugin is tried until one succeeds.

    """

    @cached_property
    def _default_tokens(self) -> TokensInterface | None:
        try:
            token_type = tokens.default
        except KeyError:
            return None
        else:
            return token_type(self.config)

    @cached_property
    def _tokens(self) -> Sequence[TokensInterface]:
        all_tokens = []
        config = self.config
        for token_type in tokens.registered.values():
            if token_type == tokens.default:
                assert self._default_tokens is not None
                all_tokens.append(self._default_tokens)
            else:
                all_tokens.append(token_type(config))
        return all_tokens

    def get_login_token(self, identifier: str, authcid: str, key: bytes, *,
                        authzid: str | None = None,
                        location: str | None = None,
                        expiration: datetime | None = None) \
            -> str | None:
        tokens = self._default_tokens
        if tokens is None:
            return None
        return tokens.get_login_token(
            identifier, authcid, key, authzid=authzid, location=location,
            expiration=expiration)

    def get_admin_token(self, admin_key: bytes | None, *,
                        authzid: str | None = None,
                        location: str | None = None,
                        expiration: datetime | None = None) \
            -> str | None:
        tokens = self._default_tokens
        if tokens is None:
            return None
        return tokens.get_admin_token(
            admin_key, authzid=authzid, location=location,
            expiration=expiration)

    def parse(self, authzid: str, token: str, *,
              admin_keys: Set[bytes] = frozenset()) \
            -> TokenCredentials:
        for tokens in self._tokens:
            try:
                return tokens.parse(authzid, token, admin_keys=admin_keys)
            except ValueError:
                pass
        raise ValueError('invalid token')




from .base import TestBase

from pymap.imap import IMAPServer


class TestStore(TestBase):

    async def test_store(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 STORE * +FLAGS (\\Seen)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Recent \\Seen))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_store_silent(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 STORE * +FLAGS.SILENT (\\Seen)\r\n')
        transport.push_write(
            b'store1 OK STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_uid_store(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 UID STORE * +FLAGS (\\Seen)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Recent \\Seen) UID 104)\r\n'
            b'store1 OK UID STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_store_add_recent(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 STORE 1 +FLAGS (\\Recent)\r\n')
        transport.push_write(
            b'* 1 FETCH (FLAGS (\\Seen))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_store_remove_recent(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 STORE * -FLAGS (\\Recent)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Recent))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_store_set_non_recent(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 STORE * FLAGS ()\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Recent))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_store_invalid(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1)
        transport.push_readline(
            b'store1 STORE * +FLAGS (\\Invalid)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Recent))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_logout()
        await self.run(transport)




from collections.abc import Mapping

import pytest
from grpclib.testing import ChannelFor
from pymapadmin.grpc.admin_grpc import SystemStub, UserStub
from pymapadmin.grpc.admin_pb2 import SUCCESS, FAILURE, UserData, \
    LoginRequest, GetUserRequest, SetUserRequest, DeleteUserRequest

from pymap.admin.handlers.system import SystemHandlers
from pymap.admin.handlers.user import UserHandlers
from pymap.interfaces.backend import BackendInterface

from .base import TestBase


class TestAdminAuth(TestBase):

    admin_token = 'MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAxNWNpZCByb2xlI' \
        'D0gYWRtaW4KMDAyZnNpZ25hdHVyZSDND4TS4f6mH9ty0DHCwqB0IIuk_IqIUFgse0OV' \
        'eHT7cAo'

    @pytest.fixture
    def overrides(self):
        return {'admin_key': b'testadmintoken'}

    async def test_token(self, backend: BackendInterface) -> None:
        token = await self._login(backend, 'testuser', 'testpass')
        await self._get_user(backend, token, 'testuser')
        await self._set_user(backend, token, 'testuser', 'newpass')
        await self._get_user(backend, token, 'testuser')
        await self._delete_user(backend, token, 'testuser')
        await self._get_user(backend, token, 'testuser',
                             failure_key='InvalidAuth')
        await self._get_user(backend, self.admin_token, 'testuser',
                             failure_key='UserNotFound')

    async def test_authorization(self, backend: BackendInterface) -> None:
        await self._set_user(backend, self.admin_token, 'newuser', 'newpass')
        token1 = await self._login(backend, 'testuser', 'testpass')
        token2 = await self._login(backend, 'newuser', 'newpass')
        await self._get_user(backend, token1, 'newuser',
                             failure_key='AuthorizationFailure')
        await self._get_user(backend, token2, 'newuser')
        await self._set_user(backend, token1, 'newuser', 'newpass2',
                             failure_key='AuthorizationFailure')
        await self._set_user(backend, token2, 'newuser', 'newpass2')
        await self._delete_user(backend, token1, 'newuser',
                                failure_key='AuthorizationFailure')
        await self._delete_user(backend, token2, 'newuser')

    async def test_admin_role(self, backend: BackendInterface) -> None:
        token = await self._login(backend, 'testuser', 'testpass')
        await self._set_user(backend, token, 'testuser', 'testpass',
                             params={'role': 'admin'},
                             failure_key='NotAllowedError')
        await self._set_user(backend, self.admin_token, 'testuser', 'testpass',
                             params={'role': 'admin'})
        await self._set_user(backend, token, 'testuser', 'testpass',
                             params={'role': 'admin'})
        await self._set_user(backend, token, 'newuser', 'newpass')
        await self._delete_user(backend, token, 'newuser')

    async def _login(self, backend, authcid: str, secret: str) -> str:
        handlers = SystemHandlers(backend)
        request = LoginRequest(authcid=authcid, secret=secret)
        async with ChannelFor([handlers]) as channel:
            stub = SystemStub(channel)
            response = await stub.Login(request)
        assert SUCCESS == response.result.code
        return response.bearer_token

    async def _get_user(self, backend: BackendInterface,
                        token: str, user: str, *,
                        failure_key: str | None = None) -> None:
        handlers = UserHandlers(backend)
        request = GetUserRequest(user=user)
        metadata = {'auth-token': token}
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.GetUser(request, metadata=metadata)
        if failure_key is None:
            assert SUCCESS == response.result.code
            assert user == response.username
        else:
            assert FAILURE == response.result.code
            assert failure_key == response.result.key

    async def _set_user(self, backend: BackendInterface,
                        token: str, user: str, password: str, *,
                        params: Mapping[str, str] | None = None,
                        failure_key: str | None = None) -> None:
        handlers = UserHandlers(backend)
        data = UserData(password=password, params=params)
        request = SetUserRequest(user=user, data=data)
        metadata = {'auth-token': token}
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.SetUser(request, metadata=metadata)
        if failure_key is None:
            assert SUCCESS == response.result.code
            assert user == response.username
        else:
            assert FAILURE == response.result.code
            assert failure_key == response.result.key

    async def _delete_user(self, backend: BackendInterface,
                           token: str, user: str, *,
                           failure_key: str | None = None) -> None:
        handlers = UserHandlers(backend)
        request = DeleteUserRequest(user=user)
        metadata = {'auth-token': token}
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.DeleteUser(request, metadata=metadata)
        if failure_key is None:
            assert SUCCESS == response.result.code
            assert user == response.username
        else:
            assert FAILURE == response.result.code
            assert failure_key == response.result.key




from textwrap import dedent

from .base import TestBase

from pymap.imap import IMAPServer


class TestMailbox(TestBase):

    async def test_list_sep(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'list1 LIST "" ""\r\n')
        transport.push_write(
            b'* LIST (\\Noselect) "/" ""\r\n'
            b'list1 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_list(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'list1 LIST "" *\r\n')
        transport.push_write(
            b'* LIST (\\HasNoChildren) "/" INBOX\r\n'
            b'* LIST (\\HasNoChildren) "/" Sent\r\n'
            b'* LIST (\\HasNoChildren) "/" Trash\r\n'
            b'list1 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_create(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'create1 CREATE "test mailbox"\r\n')
        transport.push_write(
            b'create1 OK [MAILBOXID (', (br'F[a-f0-9]+', ), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'list1 LIST "" *\r\n')
        transport.push_write(
            b'* LIST (\\HasNoChildren) "/" INBOX\r\n'
            b'* LIST (\\HasNoChildren) "/" Sent\r\n'
            b'* LIST (\\HasNoChildren) "/" Trash\r\n'
            b'* LIST (\\HasNoChildren) "/" "test mailbox"\r\n'
            b'list1 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_create_inferior(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'create1 CREATE "Trash/test mailbox"\r\n')
        transport.push_write(
            b'create1 OK [MAILBOXID (', (br'F[a-f0-9]+', ), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'list1 LIST "Trash" *\r\n')
        transport.push_write(
            b'* LIST (\\HasChildren) "/" Trash\r\n'
            b'* LIST (\\HasNoChildren) "/" "Trash/test mailbox"\r\n'
            b'list1 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_delete(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'delete1 DELETE Sent\r\n')
        transport.push_write(
            b'delete1 OK DELETE completed.\r\n')
        transport.push_readline(
            b'list2 LIST "" *\r\n')
        transport.push_write(
            b'* LIST (\\HasNoChildren) "/" INBOX\r\n'
            b'* LIST (\\HasNoChildren) "/" Trash\r\n'
            b'list2 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_delete_superior(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'create1 CREATE "Trash/test mailbox"\r\n')
        transport.push_write(
            b'create1 OK [MAILBOXID (', (br'F[a-f0-9]+', ), b')]'
            b' CREATE completed.\r\n')
        transport.push_readline(
            b'delete1 DELETE Trash\r\n')
        transport.push_write(
            b'delete1 OK DELETE completed.\r\n')
        transport.push_readline(
            b'list1 LIST "Trash" *\r\n')
        transport.push_write(
            b'* LIST (\\Noselect \\HasChildren) "/" Trash\r\n'
            b'* LIST (\\HasNoChildren) "/" "Trash/test mailbox"\r\n'
            b'list1 OK LIST completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_delete_selected(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'delete1 DELETE Sent\r\n')
        transport.push_write(
            b'* BYE Selected mailbox no longer exists.\r\n'
            b'delete1 OK DELETE completed.\r\n')
        await self.run(transport)

    async def test_lsub(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'lsub1 LSUB "" *\r\n')
        transport.push_write(
            b'* LSUB (\\HasNoChildren) "/" INBOX\r\n'
            b'lsub1 OK LSUB completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_subscribe_unsubscribe(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'subscribe1 SUBSCRIBE "Sent"\r\n')
        transport.push_write(
            b'subscribe1 OK SUBSCRIBE completed.\r\n')
        transport.push_readline(
            b'subscribe2 SUBSCRIBE "Trash"\r\n')
        transport.push_write(
            b'subscribe2 OK SUBSCRIBE completed.\r\n')
        transport.push_readline(
            b'unsubscribe1 UNSUBSCRIBE "Trash"\r\n')
        transport.push_write(
            b'unsubscribe1 OK UNSUBSCRIBE completed.\r\n')
        transport.push_readline(
            b'lsub1 LSUB "" *\r\n')
        transport.push_write(
            b'* LSUB (\\HasNoChildren) "/" INBOX\r\n'
            b'* LSUB (\\HasNoChildren) "/" Sent\r\n'
            b'lsub1 OK LSUB completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_status(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'status1 STATUS INBOX '
            b'(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN MAILBOXID)\r\n')
        transport.push_write(
            b'* STATUS INBOX (MESSAGES 4 RECENT 1 UIDNEXT 105 '
            b'UIDVALIDITY ', (br'\d+', b'uidval1'), b' UNSEEN 2 '
            b'MAILBOXID (', (br'F[a-f0-9]+', b'mbxid'), b'))\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_select(b'INBOX', 4, 1, 105, 3)
        transport.push_readline(
            b'status2 STATUS INBOX '
            b'(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n')
        transport.push_write(
            b'* STATUS INBOX (MESSAGES 4 RECENT 1 UIDNEXT 105 '
            b'UIDVALIDITY ', (br'\d+', b'uidval2'), b' UNSEEN 2)\r\n'
            b'status2 OK STATUS completed.\r\n')
        transport.push_readline(
            b'close1 CLOSE\r\n')
        transport.push_write(
            b'close1 OK CLOSE completed.\r\n')
        transport.push_readline(
            b'status3 STATUS INBOX '
            b'(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n')
        transport.push_write(
            b'* STATUS INBOX (MESSAGES 4 RECENT 0 UIDNEXT 105 '
            b'UIDVALIDITY ', (br'\d+', b'uidval2'), b' UNSEEN 2)\r\n'
            b'status3 OK STATUS completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['uidval1'] == self.matches['uidval2']
        assert self.matches['mbxid1'] == self.matches['mbxid']

    async def test_append(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        message = b'test message\r\n'
        transport.push_login()
        transport.push_readline(
            b'append1 APPEND INBOX (\\Seen) {%i}\r\n' % len(message))
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(message)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 OK [APPENDUID ', (br'\d+', ), b' 105]'
            b' APPEND completed.\r\n')
        transport.push_select(b'INBOX', 5, 2, 106, 3)
        transport.push_logout()
        await self.run(transport)

    async def test_append_empty(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'append1 APPEND INBOX {0}\r\n')
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(
            b'')
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 NO APPEND cancelled.\r\n')
        transport.push_select(b'INBOX', 4, 1, 105, 3)
        transport.push_logout()
        await self.run(transport)

    async def test_append_multi(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        message_1 = b'test message\r\n'
        message_2 = b'other test message\r\n'
        transport.push_login()
        transport.push_readline(
            b'append1 APPEND INBOX (\\Seen) {%i}\r\n' % len(message_1))
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(message_1)
        transport.push_readline(
            b' {%i}\r\n' % len(message_2))
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(message_2)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 OK [APPENDUID ', (br'\d+', ), b' 105:106]'
            b' APPEND completed.\r\n')
        transport.push_select(b'INBOX', 6, 3, 107, 3)
        transport.push_logout()
        await self.run(transport)

    async def test_append_selected(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        message = b'test message\r\n'
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, 105, 3)
        transport.push_readline(
            b'append1 APPEND INBOX (\\Seen) {%i}\r\n' % len(message))
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(message)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'* 5 EXISTS\r\n'
            b'* 2 RECENT\r\n'
            b'* 5 FETCH (FLAGS (\\Recent \\Seen))\r\n'
            b'append1 OK [APPENDUID ', (br'\d+', ), b' 105]'
            b' APPEND completed.\r\n')
        transport.push_readline(
            b'status1 STATUS INBOX (RECENT)\r\n')
        transport.push_write(
            b'* STATUS INBOX (RECENT 2)\r\n'
            b'status1 OK STATUS completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_append_email_id(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        message_1 = b'test message\r\n'
        message_2 = b'other test message\r\n'
        message_3 = message_1
        transport.push_login()
        transport.push_readline(
            b'append1 APPEND INBOX {%i+}\r\n' % len(message_1))
        transport.push_readexactly(message_1)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 OK [APPENDUID ', (br'\d+', ), b' 105]'
            b' APPEND completed.\r\n')
        transport.push_readline(
            b'append2 APPEND INBOX {%i+}\r\n' % len(message_2))
        transport.push_readexactly(message_2)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append2 OK [APPENDUID ', (br'\d+', ), b' 106]'
            b' APPEND completed.\r\n')
        transport.push_readline(
            b'append3 APPEND INBOX {%i+}\r\n' % len(message_3))
        transport.push_readexactly(message_3)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append3 OK [APPENDUID ', (br'\d+', ), b' 107]'
            b' APPEND completed.\r\n')
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 UID FETCH 105:107 (EMAILID)\r\n')
        transport.push_write(
            b'* 5 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id1'), b') '
            b'UID 105)\r\n'
            b'* 6 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id2'), b') '
            b'UID 106)\r\n'
            b'* 7 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id3'), b') '
            b'UID 107)\r\n'
            b'fetch1 OK UID FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['id1'] != self.matches['id2']
        assert self.matches['id1'] == self.matches['id3']

    async def test_append_thread_id(self, imap_server: IMAPServer) -> None:
        messages = [dedent("""\
                           Message-Id: <one>
                           Subject: thread one

                           """).encode('ascii'),
                    dedent("""\
                           Message-Id: <one>
                           Subject: Fwd: thread one

                           """).encode('ascii'),
                    dedent("""\
                           Message-Id: <one>
                           Subject: unrelated to thread one

                           """).encode('ascii'),
                    dedent("""\
                           Message-Id: <two>
                           Subject: thread two

                           """).encode('ascii'),
                    dedent("""\
                           Message-Id: <three>
                           In-Reply-To: <two>
                           Subject: Re: thread two

                           """).encode('ascii'),
                    dedent("""\
                           Message-Id: <four>
                           References: <two> <five>
                           Subject: [a list] thread two

                           """).encode('ascii'),
                    dedent("""\
                           Message-Id: <five>
                           Subject: thread two

                           """).encode('ascii')]
        transport = self.new_transport(imap_server)
        transport.push_login()
        for message in messages:
            transport.push_readline(
                b'append1 APPEND INBOX {%i+}\r\n' % len(message))
            transport.push_readexactly(message)
            transport.push_readline(
                b'\r\n')
            transport.push_write(
                b'append1 OK [APPENDUID ', (br'\d+ \d+', ), b']'
                b' APPEND completed.\r\n')
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 UID FETCH 105:* (THREADID)\r\n')
        transport.push_write(
            b'* 5 FETCH (THREADID (', (br'T[a-f0-9]+', b'id1'), b') '
            b'UID 105)\r\n'
            b'* 6 FETCH (THREADID (', (br'T[a-f0-9]+', b'id2'), b') '
            b'UID 106)\r\n'
            b'* 7 FETCH (THREADID (', (br'T[a-f0-9]+', b'id3'), b') '
            b'UID 107)\r\n'
            b'* 8 FETCH (THREADID (', (br'T[a-f0-9]+', b'id4'), b') '
            b'UID 108)\r\n'
            b'* 9 FETCH (THREADID (', (br'T[a-f0-9]+', b'id5'), b') '
            b'UID 109)\r\n'
            b'* 10 FETCH (THREADID (', (br'T[a-f0-9]+', b'id6'), b') '
            b'UID 110)\r\n'
            b'* 11 FETCH (THREADID (', (br'T[a-f0-9]+', b'id7'), b') '
            b'UID 111)\r\n'
            b'fetch1 OK UID FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert self.matches['id1'] == self.matches['id2']
        assert self.matches['id1'] != self.matches['id3']
        assert self.matches['id1'] != self.matches['id4']
        assert self.matches['id4'] == self.matches['id5']
        assert self.matches['id4'] == self.matches['id6']
        assert self.matches['id4'] == self.matches['id7']




from __future__ import annotations

import hashlib
import os.path
from collections.abc import Sequence
from importlib.resources import files
from typing import final, Generic, TypeAlias, TypeVar, Any

import msgpack
from redis.asyncio import Redis
from redis.exceptions import NoScriptError

__all__ = ['ScriptBase']

_Val: TypeAlias = int | bytes
_RetT = TypeVar('_RetT')


class ScriptBase(Generic[_RetT]):
    """The base class for redis scripts.

    Args:
        name: The script name, which should correspond to a ``.lua`` file.

    """

    __slots__ = ['_name', '_sha']

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name
        self._sha, _ = self._load()

    def _load(self) -> tuple[str, bytes]:
        fname = os.path.join('lua', f'{self._name}.lua')
        data = files(__name__).joinpath(fname).read_bytes()
        # Redis docs specify using SHA1 here:
        return hashlib.sha1(data).hexdigest(), data  # nosec

    def _convert(self, ret: Any) -> _RetT:
        converted: _RetT = ret
        return converted

    @final
    def _pack(self, val: Any) -> bytes:
        packed: bytes = msgpack.packb(val)
        return packed

    def _maybe_int(self, val: bytes) -> int | None:
        if val == b'':
            return None
        else:
            return int(val)

    async def eval(self, redis: Redis[bytes], keys: Sequence[bytes],
                   args: Sequence[_Val]) -> _RetT:
        """Run the script.

        Notes:
            This attempts the EVALSHA command first, then falls back to EVAL if
            a NOSCRIPT error is returned.

        Raises:
            :class:`~asyncio.ResponseError`

        """
        num_keys = len(keys)
        try:
            ret = await redis.evalsha(
                self._sha, num_keys, *keys, *args)  # type: ignore
        except NoScriptError:
            self._sha, data = self._load()
            ret = await redis.eval(
                data, num_keys, *keys, *args)  # type: ignore
        return self._convert(ret)




from .base import TestBase

from pymap.imap import IMAPServer


class TestFetch(TestBase):

    async def test_uid_fetch(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 UID FETCH 1:* (FLAGS)\r\n')
        transport.push_write(
            b'* 1 FETCH (FLAGS (\\Seen) UID 101)\r\n'
            b'* 2 FETCH (FLAGS (\\Answered \\Seen) UID 102)\r\n'
            b'* 3 FETCH (FLAGS (\\Flagged) UID 103)\r\n'
            b'* 4 FETCH (FLAGS (\\Recent) UID 104)\r\n'
            b'fetch1 OK UID FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_full(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH * FULL\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Recent) '
            b'INTERNALDATE "01-Jan-2010 00:00:00 +0000" '
            b'RFC822.SIZE 1980 '
            b'ENVELOPE ("01-Jan-2010 01:01:00 +0000" "Hello, World!" '
            b'(("" NIL "friend" "example.com")) '
            b'(("" NIL "friend" "example.com")) '
            b'(("" NIL "friend" "example.com")) '
            b'(("" NIL "me" "example.com")) NIL NIL NIL NIL) '
            b'BODY ("text" "plain" NIL NIL NIL "7BIT" 1980 38))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_bodystructure(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH * (BODYSTRUCTURE)\r\n')
        transport.push_write(
            b'* 4 FETCH (BODYSTRUCTURE ("text" "plain" NIL NIL NIL "7BIT" '
            b'1980 38 NIL NIL NIL NIL))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_body_section(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 3 (FLAGS)\r\n')
        transport.push_write(
            b'* 3 FETCH (FLAGS (\\Flagged))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 3 (BODY[])\r\n')
        transport.push_write(
            b'* 3 FETCH (BODY[] {205}\r\n'
            b'From: corp@example.com\r\n'
            b'To: me@example.com\r\n'
            b'Subject: Important notice regarding your account\r\n'
            b'Priority: high\r\n'
            b'Content-Type: text/plain\r\n'
            b'Date: 01-Jan-2000 01:01:00 +0000\r\n\r\n'
            b'This is some important stuff!\r\n\r\n'
            b' FLAGS (\\Flagged \\Seen))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_rfc822(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 3 (FLAGS)\r\n')
        transport.push_write(
            b'* 3 FETCH (FLAGS (\\Flagged))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 3 (RFC822)\r\n')
        transport.push_write(
            b'* 3 FETCH (RFC822 {205}\r\n'
            b'From: corp@example.com\r\n'
            b'To: me@example.com\r\n'
            b'Subject: Important notice regarding your account\r\n'
            b'Priority: high\r\n'
            b'Content-Type: text/plain\r\n'
            b'Date: 01-Jan-2000 01:01:00 +0000\r\n\r\n'
            b'This is some important stuff!\r\n\r\n'
            b' FLAGS (\\Flagged \\Seen))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_rfc822_header(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 1 (RFC822.HEADER)\r\n')
        transport.push_write(
            b'* 1 FETCH (RFC822.HEADER {142}\r\n'
            b'From: friend@example.com\r\n'
            b'To: me@example.com\r\n'
            b'Subject: Re: Re: Random question\r\n'
            b'Content-Type: text/plain\r\n'
            b'Date: 01-Jan-1970 01:01:00 +0000\r\n'
            b'\r\n)\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_rfc822_text(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 3 (FLAGS)\r\n')
        transport.push_write(
            b'* 3 FETCH (FLAGS (\\Flagged))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 3 (RFC822.TEXT)\r\n')
        transport.push_write(
            b'* 3 FETCH (RFC822.TEXT {33}\r\n'
            b'This is some important stuff!\r\n\r\n'
            b' FLAGS (\\Flagged \\Seen))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_body_section_header(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'fetch1 FETCH 2 (BODY[3.HEADER])\r\n')
        transport.push_write(
            b'* 2 FETCH (BODY[3.HEADER] {172}\r\n'
            b'From: corp@example.com\r\n'
            b'To: me@example.com\r\n'
            b'Subject: Important notice regarding your account\r\n'
            b'Priority: high\r\n'
            b'Content-Type: text/plain\r\n'
            b'Date: 01-Jan-2000 01:01:00 +0000\r\n'
            b'\r\n FLAGS (\\Draft \\Seen))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 2 (BODY[1.HEADER])\r\n')
        transport.push_write(
            b'* 2 FETCH (BODY[1.HEADER] {0}\r\n'
            b')\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_body_section_header_fields(
            self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'fetch1 FETCH 2 (BODY[3.HEADER.FIELDS (FROM TO DATE)])\r\n')
        transport.push_write(
            b'* 2 FETCH (BODY[3.HEADER.FIELDS (DATE FROM TO)] {80}\r\n'
            b'From: corp@example.com\r\n'
            b'To: me@example.com\r\n'
            b'Date: 01-Jan-2000 01:01:00 +0000\r\n'
            b'\r\n FLAGS (\\Draft \\Seen))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_body_section_text(
            self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'fetch1 FETCH 2 (FLAGS)\r\n')
        transport.push_write(
            b'* 2 FETCH (FLAGS (\\Draft))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 2 (BODY[3.TEXT])\r\n')
        transport.push_write(
            b'* 2 FETCH (BODY[3.TEXT] {33}\r\n'
            b'This is some important stuff!\r\n\r\n'
            b' FLAGS (\\Draft \\Seen))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch3 FETCH 2 (BODY[1.TEXT])\r\n')
        transport.push_write(
            b'* 2 FETCH (BODY[1.TEXT] {0}\r\n'
            b')\r\n'
            b'fetch3 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_binary_section(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'fetch1 FETCH 2 (BINARY.SIZE[1] BINARY[1])\r\n')
        transport.push_write(
            b'* 2 FETCH (BINARY.SIZE[1] 28 BINARY[1] ~{28}\r\n'
            + 'ᎻéⅬⅬՕ ᎳоᏒⅬⅮ'.encode('utf-8') +
            b' FLAGS (\\Draft \\Seen))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 2 (BINARY.SIZE[2] BINARY[2])\r\n')
        transport.push_write(
            b'* 2 FETCH (BINARY.SIZE[2] 16 BINARY[2] ~{16}\r\n'
            + 'foo=bar+ßaz\r\n\r\n'.encode('utf-8') +
            b')\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_binary_section_partial(
            self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Sent')
        transport.push_readline(
            b'fetch1 FETCH 2 (BINARY[1]<5.17>)\r\n')
        transport.push_write(
            b'* 2 FETCH (BINARY[1]<5> ~{17}\r\n'
            + 'ⅬⅬՕ ᎳоᏒ'.encode('utf-8') +
            b' FLAGS (\\Draft \\Seen))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_append_fetch_binary(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        message = b'\r\ntest\x00message\r\n'
        transport.push_login()
        transport.push_readline(
            b'append1 APPEND INBOX (\\Seen) ~{%i}\r\n' % len(message))
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(message)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 OK [APPENDUID ', (br'\d+', ), b' 105]'
            b' APPEND completed.\r\n')
        transport.push_select(b'INBOX', 5, 2, 106, 3)
        transport.push_readline(
            b'fetch1 FETCH 5 (BINARY[] BINARY.SIZE[])\r\n')
        transport.push_write(
            b'* 5 FETCH (BINARY[] ~{16}\r\n'
            b'\r\ntest\x00message\r\n'
            b' BINARY.SIZE[] 16)\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_email_id(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 1:* (EMAILID)\r\n')
        transport.push_write(
            b'* 1 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id1'), b'))\r\n'
            b'* 2 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id2'), b'))\r\n'
            b'* 3 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id3'), b'))\r\n'
            b'* 4 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id4'), b'))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 1:* (EMAILID)\r\n')
        transport.push_write(
            b'* 1 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id5'), b'))\r\n'
            b'* 2 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id6'), b'))\r\n'
            b'* 3 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id7'), b'))\r\n'
            b'* 4 FETCH (EMAILID (', (br'M[a-f0-9]+', b'id8'), b'))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)
        assert len({self.matches[f'id{n}'] for n in range(1, 5)}) == 4
        for n in range(1, 5):
            assert self.matches[f'id{n}'] == self.matches[f'id{n+4}']




import base64

from .base import TestBase

from pymap.imap import IMAPServer


class TestSession(TestBase):

    async def test_login_logout(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_logout()
        await self.run(transport)

    async def test_select(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, 105, 3)
        transport.push_logout()
        await self.run(transport)

    async def test_select_clears_recent(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, 105, 3)
        transport.push_select(b'INBOX', 4, 0, 105, 3)
        transport.push_logout()
        await self.run(transport)

    async def test_concurrent_select_clears_recent(
            self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        concurrent = self.new_transport(imap_server)
        event1, event2 = self.new_events(2)

        concurrent.push_login()
        concurrent.push_readline(
            b'status1 STATUS INBOX (MESSAGES RECENT)\r\n')
        concurrent.push_write(
            b'* STATUS INBOX (MESSAGES 4 RECENT 1)\r\n'
            b'status1 OK STATUS completed.\r\n', set=event1)
        concurrent.push_select(b'INBOX', 4, 0, 105, 3, wait=event2)
        concurrent.push_logout()

        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, 105, 3, wait=event1, set=event2)
        transport.push_logout()

        await self.run(transport, concurrent)

    async def test_auth_plain(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_write(
            b'* OK [CAPABILITY IMAP4rev1',
            (br'(?:\s+[a-zA-Z0-9=+-]+)*', ),
            b' AUTH=PLAIN',
            (br'(?:\s+[a-zA-Z0-9=+-]+)*',),
            b'] Server ready ',
            (br'\S+',), b'\r\n')
        transport.push_readline(
            b'auth1 AUTHENTICATE PLAIN\r\n')
        transport.push_write(
            b'+ \r\n')
        transport.push_readexactly(b'')
        transport.push_readline(
            base64.b64encode(b'\x00testuser\x00testpass') + b'\r\n')
        transport.push_write(
            b'auth1 OK [CAPABILITY IMAP4rev1',
            (br'(?:\s+[a-zA-Z0-9=+-]+)*', ),
            b'] Authentication successful.\r\n')
        transport.push_logout()
        await self.run(transport)




from .base import TestBase

from pymap.imap import IMAPServer


class TestExpunge(TestBase):

    async def test_expunge(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 STORE * +FLAGS (\\Deleted)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Deleted \\Recent))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_readline(
            b'expunge1 EXPUNGE\r\n')
        transport.push_write(
            b'* 4 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'expunge1 OK EXPUNGE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_expunge_uid(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'store1 UID STORE * +FLAGS (\\Deleted)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Deleted \\Recent) UID 104)\r\n'
            b'store1 OK UID STORE completed.\r\n')
        transport.push_readline(
            b'expunge1 UID EXPUNGE 101:103\r\n')
        transport.push_write(
            b'expunge1 OK UID EXPUNGE completed.\r\n')
        transport.push_readline(
            b'expunge1 UID EXPUNGE 101:*\r\n')
        transport.push_write(
            b'* 4 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'expunge1 OK UID EXPUNGE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_concurrent_expunge_responses(
            self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        concurrent = self.new_transport(imap_server)
        event1, event2 = self.new_events(2)

        concurrent.push_login()
        concurrent.push_select(b'INBOX', 4, 1, set=event1)
        concurrent.push_readline(
            b'fetch1 FETCH 1:* (UID FLAGS)\r\n', wait=event2)
        concurrent.push_write(
            b'* 1 FETCH (UID 101 FLAGS (\\Seen))\r\n'
            b'* 2 FETCH (UID 102 FLAGS (\\Answered \\Seen))\r\n'
            b'* 3 FETCH (UID 103 FLAGS (\\Flagged))\r\n'
            b'* 4 FETCH (UID 104 FLAGS (\\Deleted \\Recent))\r\n'
            b'fetch1 OK [EXPUNGEISSUED] FETCH completed.\r\n')
        concurrent.push_readline(
            b'store2 STORE 1:* +FLAGS (\\Flagged)\r\n')
        concurrent.push_write(
            b'* 1 FETCH (FLAGS (\\Flagged \\Seen))\r\n'
            b'* 2 FETCH (FLAGS (\\Answered \\Flagged \\Seen))\r\n'
            b'* 3 FETCH (FLAGS (\\Flagged))\r\n'
            b'* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Recent))\r\n'
            b'store2 OK [EXPUNGEISSUED] STORE completed.\r\n')
        concurrent.push_readline(
            b'search1 SEARCH ALL\r\n')
        concurrent.push_write(
            b'* SEARCH 1 2 3 4\r\n'
            b'search1 OK [EXPUNGEISSUED] SEARCH completed.\r\n')
        concurrent.push_readline(
            b'noop1 NOOP\r\n')
        concurrent.push_write(
            b'* 4 EXPUNGE\r\n'
            b'* 0 RECENT\r\n'
            b'noop1 OK NOOP completed.\r\n')
        concurrent.push_logout()

        transport.push_login()
        transport.push_select(b'INBOX', 4, 0, wait=event1)
        transport.push_readline(
            b'store1 STORE * +FLAGS (\\Deleted)\r\n')
        transport.push_write(
            b'* 4 FETCH (FLAGS (\\Deleted))\r\n'
            b'store1 OK STORE completed.\r\n')
        transport.push_readline(
            b'expunge1 EXPUNGE\r\n', set=event2)
        transport.push_write(
            b'* 4 EXPUNGE\r\n'
            b'expunge1 OK EXPUNGE completed.\r\n')
        transport.push_logout()

        await self.run(transport, concurrent)




from .base import TestBase

from pymap.imap import IMAPServer


class TestIdle(TestBase):

    async def test_idle(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, 105)
        transport.push_readline(
            b'idle1 IDLE\r\n')
        transport.push_write(
            b'+ Idling.\r\n')
        transport.push_readexactly(b'')
        transport.push_readline(
            b'DONE\r\n')
        transport.push_write(
            b'idle1 OK IDLE completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_idle_invalid(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, 105)
        transport.push_readline(
            b'idle1 IDLE\r\n')
        transport.push_write(
            b'+ Idling.\r\n')
        transport.push_readexactly(b'')
        transport.push_readline(
            b'WHAT\r\n')
        transport.push_write(
            b'idle1 BAD Expected "DONE".\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_idle_noselect(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_readline(
            b'idle1 IDLE\r\n')
        transport.push_write(
            b'idle1 BAD IDLE: Must select a mailbox first.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_concurrent_idle_append(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        concurrent = self.new_transport(imap_server)
        event1, event2, event3 = self.new_events(3)

        concurrent.push_login()
        concurrent.push_select(b'INBOX', 4, 1, 105)
        concurrent.push_readline(
            b'idle1 IDLE\r\n')
        concurrent.push_write(
            b'+ Idling.\r\n')
        concurrent.push_readexactly(b'', set=event1, wait=event3)
        concurrent.push_write(
            b'* 5 EXISTS\r\n')
        concurrent.push_write(
            b'* 2 RECENT\r\n')
        concurrent.push_write(
            b'* 5 FETCH (FLAGS (\\Recent \\Seen))\r\n', set=event3)
        concurrent.push_readline(
            b'DONE\r\n')
        concurrent.push_write(
            b'idle1 OK IDLE completed.\r\n')
        concurrent.push_logout()

        transport.push_login()
        transport.push_readline(
            b'append1 APPEND INBOX (\\Seen) {9}\r\n', wait=event1)
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(
            b'testing\r\n')
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 OK [APPENDUID ', None, b' 105] APPEND completed.\r\n',
            set=event2)
        transport.push_logout()

        await self.run(transport, concurrent)

    async def test_concurrent_idle_expunge(self, imap_server: IMAPServer) \
            -> None:
        transport = self.new_transport(imap_server)
        concurrent = self.new_transport(imap_server)
        event1, event2, event3 = self.new_events(3)

        concurrent.push_login()
        concurrent.push_select(b'INBOX', 4, 1, 105)
        concurrent.push_readline(
            b'idle1 IDLE\r\n')
        concurrent.push_write(
            b'+ Idling.\r\n')
        concurrent.push_readexactly(b'', set=event1, wait=event3)
        concurrent.push_write(
            b'* 1 EXPUNGE\r\n', set=event3)
        concurrent.push_readline(
            b'DONE\r\n', wait=event2)
        concurrent.push_write(
            b'idle1 OK IDLE completed.\r\n')
        concurrent.push_logout()

        transport.push_login()
        transport.push_select(b'INBOX', 4, 0, 105, wait=event1)
        transport.push_readline(
            b'store1 STORE 1 +FLAGS.SILENT (\\Deleted)\r\n')
        transport.push_write(
            b'store1 OK STORE completed.\r\n')
        transport.push_readline(
            b'close1 CLOSE\r\n')
        transport.push_write(
            b'close1 OK CLOSE completed.\r\n', set=event2)
        transport.push_logout()

        await self.run(transport, concurrent)




import pytest
from grpclib.testing import ChannelFor
from pymapadmin.grpc.admin_grpc import UserStub
from pymapadmin.grpc.admin_pb2 import SUCCESS, FAILURE, \
    GetUserRequest, SetUserRequest, DeleteUserRequest, UserData

from pymap.admin.handlers.user import UserHandlers
from pymap.imap import IMAPServer
from pymap.interfaces.backend import BackendInterface

from .base import TestBase


class TestMailboxHandlers(TestBase):

    admin_token = 'MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAxNWNpZCByb2xlI' \
        'D0gYWRtaW4KMDAyZnNpZ25hdHVyZSDND4TS4f6mH9ty0DHCwqB0IIuk_IqIUFgse0OV' \
        'eHT7cAo'
    metadata = {'auth-token': admin_token}

    @pytest.fixture
    def overrides(self):
        return {'admin_key': b'testadmintoken'}

    async def test_get_user(self, backend: BackendInterface) -> None:
        handlers = UserHandlers(backend)
        request = GetUserRequest(user='testuser')
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.GetUser(request, metadata=self.metadata)
        assert SUCCESS == response.result.code
        assert 'testuser' == response.username
        assert '$pbkdf2$1$$FzEpdTtdOaIFkUucxV4PjfW88BE=' \
            == response.data.password

    async def test_get_user_not_found(self, backend: BackendInterface) -> None:
        handlers = UserHandlers(backend)
        request = GetUserRequest(user='baduser')
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.GetUser(request, metadata=self.metadata)
        assert FAILURE == response.result.code
        assert 'UserNotFound' == response.result.key

    async def test_set_user(self, backend: BackendInterface,
                            imap_server: IMAPServer) -> None:
        handlers = UserHandlers(backend)
        data = UserData(password='newpass', params={'key': 'val'})
        request = SetUserRequest(user='testuser', data=data)
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.SetUser(request, metadata=self.metadata)
        assert SUCCESS == response.result.code
        assert 'testuser' == response.username

        transport = self.new_transport(imap_server)
        transport.push_login(password=b'newpass')
        transport.push_select(b'INBOX')
        transport.push_logout()
        await self.run(transport)

    async def test_delete_user(self, backend: BackendInterface) -> None:
        handlers = UserHandlers(backend)
        request = DeleteUserRequest(user='testuser')
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.DeleteUser(request, metadata=self.metadata)
        assert SUCCESS == response.result.code
        assert 'testuser' == response.username

    async def test_delete_user_not_found(self, backend: BackendInterface) \
            -> None:
        handlers = UserHandlers(backend)
        request = DeleteUserRequest(user='baduser')
        async with ChannelFor([handlers]) as channel:
            stub = UserStub(channel)
            response = await stub.DeleteUser(request, metadata=self.metadata)
        assert FAILURE == response.result.code
        assert 'UserNotFound' == response.result.key




from .base import TestBase

from pymap.imap import IMAPServer


class TestReadOnly(TestBase):

    async def test_select(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Trash', 1, 1, readonly=True)
        transport.push_select(b'Trash', 1, 1, readonly=True)
        transport.push_logout()
        await self.run(transport)

    async def test_examine(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX', 4, 1, examine=True)
        transport.push_select(b'INBOX', 4, 1, examine=True)
        transport.push_logout()
        await self.run(transport)

    async def test_append(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        message = b'test message\r\n'
        transport.push_login()
        transport.push_readline(
            b'append1 APPEND Trash (\\Seen) {%i}\r\n' % len(message))
        transport.push_write(
            b'+ Literal string\r\n')
        transport.push_readexactly(message)
        transport.push_readline(
            b'\r\n')
        transport.push_write(
            b'append1 NO [READ-ONLY] Mailbox is read-only.\r\n')
        transport.push_select(b'Trash', 1, readonly=True)
        transport.push_logout()
        await self.run(transport)

    async def test_copy(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'copy1 COPY 1 Trash\r\n')
        transport.push_write(
            b'copy1 NO [READ-ONLY] Mailbox is read-only.\r\n')
        transport.push_select(b'Trash', 1, readonly=True)
        transport.push_logout()
        await self.run(transport)

    async def test_expunge(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Trash', 1, readonly=True)
        transport.push_readline(
            b'expunge1 EXPUNGE\r\n')
        transport.push_write(
            b'expunge1 NO [READ-ONLY] Mailbox is read-only.\r\n')
        transport.push_select(b'Trash', 1, readonly=True)
        transport.push_logout()
        await self.run(transport)

    async def test_store(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Trash', 1, readonly=True)
        transport.push_readline(
            b'store1 STORE 1 +FlAGS (\\Seen)\r\n')
        transport.push_write(
            b'store1 NO [READ-ONLY] Mailbox is read-only.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_fetch_not_seen(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'Trash', 1, readonly=True)
        transport.push_readline(
            b'fetch1 FETCH 1 (FLAGS)\r\n')
        transport.push_write(
            b'* 1 FETCH (FLAGS (\\Deleted))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'fetch2 FETCH 1 (RFC822.TEXT FLAGS)\r\n')
        transport.push_write(
            b'* 1 FETCH (RFC822.TEXT {53}\r\n'
            b'It just works. Only five easy payments of $19.99.\r\n\r\n'
            b' FLAGS (\\Deleted))\r\n'
            b'fetch2 OK FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)




import pytest
from grpclib.testing import ChannelFor
from pymapadmin.grpc.admin_grpc import MailboxStub
from pymapadmin.grpc.admin_pb2 import AppendRequest, SUCCESS, FAILURE

from pymap.admin.handlers.mailbox import MailboxHandlers
from pymap.imap import IMAPServer
from pymap.interfaces.backend import BackendInterface

from .base import TestBase


class TestMailboxHandlers(TestBase):

    admin_token = 'MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAxNWNpZCByb2xlI' \
        'D0gYWRtaW4KMDAyZnNpZ25hdHVyZSDND4TS4f6mH9ty0DHCwqB0IIuk_IqIUFgse0OV' \
        'eHT7cAo'
    metadata = {'auth-token': admin_token}

    @pytest.fixture
    def overrides(self):
        return {'admin_key': b'testadmintoken'}

    async def test_append(self, backend: BackendInterface,
                          imap_server: IMAPServer) -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: user@example.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert SUCCESS == response.result.code
        assert 105 == response.uid

        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 UID FETCH * FULL\r\n')
        transport.push_write(
            b'* 5 FETCH (FLAGS (\\Flagged \\Recent \\Seen)'
            b' INTERNALDATE "13-Feb-2009 23:31:30 +0000"'
            b' RFC822.SIZE 38'
            b' ENVELOPE (NIL NIL (("" NIL "user" "example.com"))'
            b' (("" NIL "user" "example.com")) (("" NIL "user" "example.com"))'
            b' NIL NIL NIL NIL NIL)'
            b' BODY ("text" "plain" NIL NIL NIL "7BIT" 38 3)'
            b' UID 105)\r\n'
            b'fetch1 OK UID FETCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_append_user_not_found(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        request = AppendRequest(user='baduser')
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert FAILURE == response.result.code
        assert 'UserNotFound' == response.result.key

    async def test_append_mailbox_not_found(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        request = AppendRequest(user='testuser', mailbox='BAD')
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert FAILURE == response.result.code
        assert 'BAD' == response.mailbox
        assert 'MailboxNotFound' == response.result.key

    async def test_append_filter_reject(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        data = b'Subject: reject this\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert FAILURE == response.result.code
        assert 'AppendFailure' == response.result.key

    async def test_append_filter_discard(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        data = b'Subject: discard this\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert SUCCESS == response.result.code
        assert not response.mailbox
        assert not response.uid

    async def test_append_filter_address_is(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: foo@example.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 1' == response.mailbox

    async def test_append_filter_address_contains(
            self, backend: BackendInterface) -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: user@foo.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 2' == response.mailbox

    async def test_append_filter_address_matches(
            self, backend: BackendInterface) -> None:
        handlers = MailboxHandlers(backend)
        data = b'To: bigfoot@example.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 3' == response.mailbox

    async def test_append_filter_envelope_is(
            self, backend: BackendInterface) -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: user@example.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                sender='foo@example.com',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 4' == response.mailbox

    async def test_append_filter_envelope_contains(
            self, backend: BackendInterface) -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: user@example.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                sender='user@foo.com',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 5' == response.mailbox

    async def test_append_filter_envelope_matches(
            self, backend: BackendInterface) -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: user@example.com\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                recipient='bigfoot@example.com',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 6' == response.mailbox

    async def test_append_filter_exists(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        data = b'X-Foo: foo\nX-Bar: bar\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 7' == response.mailbox

    async def test_append_filter_header(self, backend: BackendInterface) \
            -> None:
        handlers = MailboxHandlers(backend)
        data = b'X-Caffeine: C8H10N4O2\n\ntest message!\n'
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 8' == response.mailbox

    async def test_append_filter_size(self, backend: BackendInterface) -> None:
        handlers = MailboxHandlers(backend)
        data = b'From: user@example.com\n\ntest message!\n'
        data = data + b'x' * (1234 - len(data))
        request = AppendRequest(user='testuser', mailbox='INBOX',
                                flags=['\\Flagged', '\\Seen'],
                                when=1234567890, data=data)
        async with ChannelFor([handlers]) as channel:
            stub = MailboxStub(channel)
            response = await stub.Append(request, metadata=self.metadata)
        assert 'Test 9' == response.mailbox




import pytest

from .base import TestBase

from pymap.imap import IMAPServer


class TestSearch(TestBase):

    @pytest.fixture
    def overrides(self):
        return {'disable_search_keys': [b'DRAFT']}

    async def test_search_disabled(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH DRAFT\r\n')
        transport.push_write(
            b'search1 NO [CANNOT] SEARCH DRAFT not supported.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH ALL\r\n')
        transport.push_write(
            b'* SEARCH 1 2 3 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_not(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH NOT ALL\r\n')
        transport.push_write(
            b'* SEARCH\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_uid(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 UID SEARCH ALL\r\n')
        transport.push_write(
            b'* SEARCH 101 102 103 104\r\n'
            b'search1 OK UID SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_seqset(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH 2:3\r\n')
        transport.push_write(
            b'* SEARCH 2 3\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_and(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH 1:3 2:4\r\n')
        transport.push_write(
            b'* SEARCH 2 3\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_or(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH OR 1 3:4\r\n')
        transport.push_write(
            b'* SEARCH 1 3 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_seen(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH SEEN\r\n')
        transport.push_write(
            b'* SEARCH 1 2\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_unseen(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH UNSEEN\r\n')
        transport.push_write(
            b'* SEARCH 3 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_new(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH NEW\r\n')
        transport.push_write(
            b'* SEARCH 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_date_on(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH ON 01-Jan-1990\r\n')
        transport.push_write(
            b'* SEARCH 2\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_date_since(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH SINCE 01-Jan-2000\r\n')
        transport.push_write(
            b'* SEARCH 3 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_header_date(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH SENTON 01-Jan-1990\r\n')
        transport.push_write(
            b'* SEARCH 2\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_size(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH SMALLER 1000\r\n')
        transport.push_write(
            b'* SEARCH 1 2 3\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_from(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH FROM "corp@example.com"\r\n')
        transport.push_write(
            b'* SEARCH 3\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_subject(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH SUBJECT "Hello"\r\n')
        transport.push_write(
            b'* SEARCH 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_header(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH HEADER "Priority" "high"\r\n')
        transport.push_write(
            b'* SEARCH 3\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_body(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH BODY "velocity"\r\n')
        transport.push_write(
            b'* SEARCH 2\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_text(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'search1 SEARCH TEXT "WORLD"\r\n')
        transport.push_write(
            b'* SEARCH 4\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_emailid(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 1,2 (EMAILID)\r\n')
        transport.push_write(
            b'* 1 FETCH (EMAILID (', (b'M[a-f0-9]+', b'id1'), b'))\r\n'
            b'* 2 FETCH (EMAILID (', (b'M[a-f0-9]+', b'id2'), b'))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'search1 SEARCH EMAILID %(id1)b\r\n')
        transport.push_write(
            b'* SEARCH 1\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_readline(
            b'search2 SEARCH EMAILID %(id2)b\r\n')
        transport.push_write(
            b'* SEARCH 2\r\n'
            b'search2 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)

    async def test_search_threadid(self, imap_server: IMAPServer) -> None:
        transport = self.new_transport(imap_server)
        transport.push_login()
        transport.push_select(b'INBOX')
        transport.push_readline(
            b'fetch1 FETCH 1,2 (THREADID)\r\n')
        transport.push_write(
            b'* 1 FETCH (THREADID (', (b'T[a-f0-9]+', b'id1'), b'))\r\n'
            b'* 2 FETCH (THREADID (', (b'T[a-f0-9]+', b'id2'), b'))\r\n'
            b'fetch1 OK FETCH completed.\r\n')
        transport.push_readline(
            b'search1 SEARCH THREADID %(id1)b\r\n')
        transport.push_write(
            b'* SEARCH 1\r\n'
            b'search1 OK SEARCH completed.\r\n')
        transport.push_readline(
            b'search2 SEARCH THREADID %(id2)b\r\n')
        transport.push_write(
            b'* SEARCH 2\r\n'
            b'search2 OK SEARCH completed.\r\n')
        transport.push_logout()
        await self.run(transport)




import asyncio
from argparse import Namespace
from asyncio import StreamReader, StreamWriter
from collections.abc import Iterable

import pytest
from proxyprotocol.sock import SocketInfoLocal
from pysasl.hashing import BuiltinHash

from pymap.backend.dict import DictBackend
from pymap.concurrent import Event
from pymap.context import subsystem
from pymap.imap import IMAPServer
from pymap.sieve.manage import ManageSieveServer

from .mocktransport import MockTransport


class FakeArgs(Namespace):

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.__dict__.update(kwargs)

    debug = True
    demo_data = 'pymap.backend.dict'
    demo_user = 'testuser'
    demo_password = 'testpass'

    def __getattr__(self, key: str) -> None:
        return None


class TestBase:

    # For speed and determinism.
    _hash_context = BuiltinHash(hash_name='sha1', salt_len=0, rounds=1)

    @classmethod
    @pytest.fixture(autouse=True)
    def init(cls, request, backend):
        test = request.instance
        test._fd = 1
        test._matches = {}

    @pytest.fixture
    def imap_server(self, backend):
        return IMAPServer(backend.login, backend.config)

    @pytest.fixture
    def managesieve_server(self, backend):
        return ManageSieveServer(backend.login, backend.config)

    @pytest.fixture
    def args(self):
        return FakeArgs()

    @pytest.fixture
    def overrides(self):
        return {}

    @pytest.fixture
    async def backend(self, args, overrides):
        backend, config = await DictBackend.init(
            args,
            hash_context=self._hash_context,
            invalid_user_sleep=0.0,
            **overrides)
        return backend

    def _incr_fd(self):
        fd = self._fd
        self._fd += 1
        return fd

    @property
    def matches(self) -> dict[str, bytes]:
        return self._matches  # type: ignore

    def new_transport(self, server: IMAPServer | ManageSieveServer) \
            -> MockTransport:
        return MockTransport(server, self.matches, self._incr_fd())

    def new_events(self, n: int) -> Iterable[Event]:
        return (subsystem.get().new_event() for _ in range(n))

    def _check_queue(self, transport: MockTransport) -> None:
        queue = transport.queue
        assert 0 == len(queue), 'Items left on queue: ' + repr(queue)

    async def _run_transport(self, transport: MockTransport) -> None:
        server = transport.server
        reader: StreamReader = transport  # type: ignore
        writer: StreamWriter = transport  # type: ignore
        sock_info = SocketInfoLocal(transport)
        return await server(reader, writer, sock_info)

    async def run(self, *transports: MockTransport) -> None:
        failures = []
        transport_tasks = [asyncio.create_task(
            self._run_transport(transport)) for transport in transports]
        for task in transport_tasks:
            try:
                await task
            except Exception as exc:
                failures.append(exc)
        if failures:
            raise failures[0]
        for transport in transports:
            self._check_queue(transport)




from grpclib.testing import ChannelFor
from pymapadmin.__about__ import __version__ as pymap_admin_version
from pymapadmin.grpc.admin_grpc import SystemStub
from pymapadmin.grpc.admin_pb2 import SUCCESS, FAILURE, \
    LoginRequest, PingRequest

from pymap.__about__ import __version__ as pymap_version
from pymap.admin.handlers.system import SystemHandlers
from pymap.interfaces.backend import BackendInterface

from .base import TestBase


class TestSystemHandlers(TestBase):

    async def test_login(self, backend: BackendInterface) -> None:
        handlers = SystemHandlers(backend)
        request = LoginRequest(authcid='testuser', secret='testpass')
        async with ChannelFor([handlers]) as channel:
            stub = SystemStub(channel)
            response = await stub.Login(request)
        assert SUCCESS == response.result.code
        assert response.bearer_token

    async def test_login_failure(self, backend: BackendInterface) -> None:
        handlers = SystemHandlers(backend)
        request = LoginRequest(authcid='baduser', secret='badpass')
        async with ChannelFor([handlers]) as channel:
            stub = SystemStub(channel)
            response = await stub.Login(request)
        assert FAILURE == response.result.code
        assert 'InvalidAuth' == response.result.key
        assert not response.HasField('bearer_token')

    async def test_ping(self, backend: BackendInterface) -> None:
        handlers = SystemHandlers(backend)
        request = PingRequest()
        async with ChannelFor([handlers]) as channel:
            stub = SystemStub(channel)
            response = await stub.Ping(request)
        assert SUCCESS == response.result.code
        assert pymap_version == response.pymap_version
        assert pymap_admin_version == response.pymap_admin_version




from .base import TestBase
from .mocktransport import MockTransport

from pymap.sieve.manage import ManageSieveServer


class TestManageSieve(TestBase):

    def _push_capabilities(self, transport: MockTransport) -> None:
        transport.push_write(
            b'"IMPLEMENTATION" "pymap managesieve', (br'.*?', ), b'"\r\n'
            b'"SASL" "PLAIN LOGIN"\r\n'
            b'"SIEVE" "fileinto reject envelope body"\r\n'
            b'"UNAUTHENTICATE"\r\n'
            b'"VERSION" "1.0"\r\n'
            b'OK\r\n')

    def _push_logout(self, transport: MockTransport) -> None:
        transport.push_readline(
            b'logout\r\n')
        transport.push_write(
            b'BYE\r\n')

    def _push_authenticate(self, transport: MockTransport) -> None:
        transport.push_readline(
            b'authenticate "plain"\r\n')
        transport.push_write(
            b'""\r\n')
        transport.push_readline(
            b'{24+}\r\n'
            b'AHRlc3R1c2VyAHRlc3RwYXNz\r\n')
        transport.push_write(
            b'OK\r\n')

    async def test_capabilities(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_logout(transport)
        await self.run(transport)

    async def test_authenticate(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        self._push_logout(transport)
        await self.run(transport)

    async def test_capability(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        transport.push_readline(
            b'capability\r\n')
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'capability\r\n')
        transport.push_write(
            b'"IMPLEMENTATION" "pymap managesieve', (br'.*?', ), b'"\r\n'
            b'"SIEVE" "fileinto reject envelope body"\r\n'
            b'"OWNER" "testuser"\r\n'
            b'"UNAUTHENTICATE"\r\n'
            b'"VERSION" "1.0"\r\n'
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_unauthenticate(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'unauthenticate\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'capability\r\n')
        self._push_capabilities(transport)
        self._push_logout(transport)
        await self.run(transport)

    async def test_bad_command(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        transport.push_readline(
            b'bad\r\n')
        transport.push_write(
            b'NO "Bad command: [:ERROR:]bad"\r\n')
        self._push_authenticate(transport)
        transport.push_readline(
            b'bad again\r\n')
        transport.push_write(
            b'NO "Bad command: [:ERROR:]bad again"\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_listscripts(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'listscripts\r\n')
        transport.push_write(
            b'"demo" ACTIVE\r\n'
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_getscript(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'getscript "demo"\r\n')
        transport.push_write(
            b'{937}\r\nrequire ["fileinto", "envelope", "reject"];\n\n'
            b'if header :is "Subject" "reject this" {\n    '
            b'reject "message rejected";\n}\n\n'
            b'if header :is "Subject" "discard this" {\n    '
            b'discard;\n}\n\n'
            b'if address :is :all "from" "foo@example.com" {\n    '
            b'fileinto "Test 1";\n}\n\n'
            b'if address :contains :domain "from" "foo" {\n    '
            b'fileinto "Test 2";\n}\n\n'
            b'if address :matches :localpart "to" "*foo?" {\n    '
            b'fileinto "Test 3";\n}\n\n'
            b'if envelope :is :all "from" "foo@example.com" {\n    '
            b'fileinto "Test 4";\n}\n\n'
            b'if envelope :contains :domain "from" "foo" {\n    '
            b'fileinto "Test 5";\n}\n\n'
            b'if envelope :matches :localpart "to" "*foo?" {\n    '
            b'fileinto "Test 6";\n}\n\n'
            b'if exists ["X-Foo", "X-Bar"] {\n    '
            b'fileinto "Test 7";\n}\n\n'
            b'if header :is ["X-Caffeine"] ["C8H10N4O2"] {\n    '
            b'fileinto "Test 8";\n}\n\n'
            b'if allof(not size :under 1234, not size :over 1234) {\n    '
            b'fileinto "Test 9";\n}\n\n'
            b'if allof (true, false) {\n    discard;\n} '
            b'elsif false {\n    discard;\n} '
            b'elsif not true {\n    discard;\n} '
            b'else {\n    keep;\n}\n\r\n'
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_havespace(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'havespace "demo" 1234\r\n')
        transport.push_write(
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_putscript(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'putscript "test" {10+}\r\n'
            b'discard;\r\n\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'setactive "test"\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'listscripts\r\n')
        transport.push_write(
            b'"demo"\r\n'
            b'"test" ACTIVE\r\n'
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_deletescript(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'deletescript "demo"\r\n')
        transport.push_write(
            b'NO (ACTIVE)\r\n')
        transport.push_readline(
            b'setactive ""\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'deletescript "demo"\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'listscripts\r\n')
        transport.push_write(
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_renamescript(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'renamescript "demo" "test"\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'listscripts\r\n')
        transport.push_write(
            b'"test" ACTIVE\r\n'
            b'OK\r\n')
        self._push_logout(transport)
        await self.run(transport)

    async def test_checkscript(
            self, managesieve_server: ManageSieveServer) -> None:
        transport = self.new_transport(managesieve_server)
        self._push_capabilities(transport)
        self._push_authenticate(transport)
        transport.push_readline(
            b'checkscript {10+}\r\n'
            b'discard;\r\n\r\n')
        transport.push_write(
            b'OK\r\n')
        transport.push_readline(
            b'checkscript {10+}\r\n'
            b'1234567890\r\n')
        transport.push_write(
            b'NO {67}\r\nline 1: parsing error: unexpected token '
            b"'1234567890' found near '1'\r\n")
        transport.push_readline(
            b'checkscript {19+}\r\n'
            b'else { discard; }\r\n\r\n')
        transport.push_write(
            b'NO {74}\r\n'
            b'line 1: parsing error: the else command must follow an '
            b'if or elsif command\r\n')
        self._push_logout(transport)
        await self.run(transport)




from __future__ import annotations

from asyncio import AbstractEventLoop

from grpclib.server import Server as _Server

__all__ = ['Server']


class Server(_Server):
    """The :class:`~grpclib.server.Server` class included with :mod:`grpclib`
    has typing issues due to some missing methods. These methods do not seem to
    be part of its public API, so they should be implemented to simply raise
    exceptions.

    Note:
        If this is fixed upstream, this class should be removed.

    """

    def get_loop(self) -> AbstractEventLoop:
        raise NotImplementedError()

    def is_serving(self) -> bool:
        raise NotImplementedError()

    async def serve_forever(self) -> None:
        raise NotImplementedError()

    async def start_serving(self) -> None:
        raise NotImplementedError()
